#!/usr/bin/env perl
=cut
Copyright (c) 2005 Albert van der Horst <albert@spenarnc.xs4all.nl>
$Id: ass.pl,v 1.24 2005/01/23 02:22:42 albert Exp $

This is an assembler according to the postit-fixup principle.
It consists of a small engine that interprets the tables, to
generate 1,2 and 3 byte opcodes and fixups from reverse.
Then the assembler file is interpreted where elements are divided
into keywords (found in the tables), numbers, labels and commaers.
This assembler doesn't understand the 16 bits operands that go
with an AS: prefix, they are not tested with the testset386.

This whole is fairly generic, except that it assumes
THE TARGET PROCESSOR IS LITTLE-ENDIAN.

For details see
    http://home.hccnet.nl/a.w.m.van.der.horst/forthassembler.html
notably the pdf's of ciforth.

=cut

use strict;

# Numbers remembered for a short while.
# The magic number is there to detect stack underflow.
our @numbers = ( 0xDEADBEEF );
#our $sp=0;            # "stack pointer"

# The code output, byte by byte. Don't use pop() and push(),
# That will be inconvenient when we add a second pass.
our @code = ( );
our $cp;              # "code pointer" 

# The following two hashes are "parallel", meaning that the hash index
# retrieves two properties of the same object. This was require with Perl 4.
# With Perl 5 we could have one hash table with a list value.
# For the moment I don't want to assume everybody runs Perl 5.

# Hash for opcodes. Constant after the table load.
our %opcode = ( );
keys(%opcode) = 1024;
our $os=0;

# Hash for opcode length. Constant after the table load.
# Find the machine code that goes with a mnemonic.
our %opcode_length = ( );
keys(%opcode_length) = 1024;

# Hash for fixups. Constant after the table load.
# Find the fixup code that goes with a mnemonic.
our %fixup = ( );
keys(%fixup) = 64;

# Hash for commaers. Find the length in bytes of the commaer-mnemonic.
our %commaer =
(
    '(C,)' , 1 ,
    '(W,)' , 2 ,
    '(L,)' , 4 ,
);

sub dump_numbers
{
    print " --------numbers--------------\n";
    print "stack depth =", $#numbers,"\n";
    if ( $#numbers < 0 ) { die "Underflow in numbers\n"; }
    foreach( @numbers ) { print $_,"\n"; }
    print " -----------------------------\n";
}

# Open all files.
sub get_file_handles
{
    if ($#ARGV != 1 )
    {
        die "Assembler for 80386\nUsage: ass.pl <INFILE> <OUTFILE>\n";
    }
    open(INFILE,  "$ARGV[0]" ) or die "Input file $ARGV[0], error: $!\n";
    open(OUTFILE, ">$ARGV[1]" ) or die "Output file $ARGV[1], error: $!\n";  ;
}


# Remember a hex number.
sub hex_number
{
    my @input=@_;
    my $number = shift(@input);
    $number =~ s/,//g;  # Get rid of the ciforth ephelkusticon.
    push( @numbers, hex($number));   
    return @input;
}

sub return_number
{
    return pop(@numbers);
}

#######################################################################
#########################                     #########################
#########################  READ TABLES        #########################
#########################                     #########################
#######################################################################

# All add_xxx function accept a (part of a ) line, starting with a keyword
# then parse and return the part not yet parsed. Sometimes also a length.
# This is the number of bytes of the opcode.

# Associate opcode with the latest number. Keyword 1pi, 2pi, 3pi,
sub add_opcode  # (length, remainder of line)
{
    my @input=@_;
    my $length = shift(@input);
    shift(@input);      # Rid of #pi.
    my $mnemonic = shift(@input);
    if ( !  $mnemonic ) { die "zero length name!\n";}
    $opcode{ $mnemonic } = return_number ;
    $opcode_length{ $mnemonic } = $length;
    return @input;
}

# Associate fixup with the latest number. Keyword fir.
sub add_fixup           # (remainder of line)
{
    my @input=@_;
    shift(@input);      # Rid of fir
    $fixup{ shift(@input) } = return_number ;
    return @input;
}

# Expect a certain string or die, (belongs of table read in)
# A mistake in the tables, really is a fatal bug in the assembler itself.
sub expect
{
    my @input=@_;
    my $required = shift(@input);
    if ( $required ne shift(@input) )
    {
        die("Error in tables! alias expected at: @input\n");
    }
    return @input;
}

# Associate commaer after "alias" with the same length as the one
# before the alias. # This is triggered by tick: '\''.
sub add_commaer         # (remainder of line)
{
    my @input=@_;
    shift(@input);      # Rid of tick.
    my $length = $commaer{shift(@input)};
    @input = expect( "alias", @input );
    $commaer{ shift(@input) } = $length ;
    return @input;
}

# Add_family_xx adds a number of opcodes/fixups at the same time.

# Associate ``length'' opcodes with the latest number plus multiple of
# the next number.
sub add_family  # (length, remainder of line)
{
    my @input=@_;
    my $length = shift(@input);
    shift(@input);      # Rid of #family,
    my $count = return_number;
    my $machinecode = return_number;
    my $increment = return_number;
    for ( 1..$count )
    {
        my $mnemonic = shift(@input);
        if ( !  $mnemonic ) { die "zero length name!\n";}
        $opcode{ $mnemonic } = $machinecode;
        $opcode_length{ $mnemonic } = $length;
        $machinecode += $increment;
    }
    if ( $#numbers != 0 ) { die "Numbers not correctly consumed!\n"; }
    return @input;
}

# Associate ``length'' fixups with the latest number plus multiple of
# the next number.
sub add_family_fix  # (remainder of line)
{
    my @input=@_;
    shift(@input);      # Rid of family|r
    my $count = return_number;
    my $machinecode = return_number;
    my $increment = return_number;
    for ( 1..$count )
    {
        my $mnemonic = shift(@input);
        if ( !  $mnemonic ) { die "zero length name!\n";}
        $fixup{ $mnemonic } = $machinecode;
        $machinecode += $increment;
    }
    if ( $#numbers != 0 ) { die "Numbers not correctly consumed!\n"; }
    return @input;
}

sub get_tables
{
    open(TABLE,  "asi386.pl" );

    while(<TABLE>)
    {
        next if /^$/;

        my @pe = split(' ',$_);

        # Within the table ``\ '' is the (Forth-compatible) comment.
        while(exists $pe[0] and $pe[0] ne "\\")
        {
            if ( $pe[0] =~ /^[0-9A-F,]*$/ )
            {
                @pe = hex_number( @pe );
            }
            elsif ( $pe[0] eq "1pi" ) { @pe = add_opcode( 1, @pe ); }
            elsif ( $pe[0] eq "2pi" ) { @pe = add_opcode( 2, @pe ); }
            elsif ( $pe[0] eq "3pi" ) { @pe = add_opcode( 3, @pe ); }
            elsif ( $pe[0] eq "fir" ) { @pe = add_fixup( @pe ); }
            elsif ( $pe[0] eq "1family," ) { @pe = add_family( 1, @pe ); }
            elsif ( $pe[0] eq "2family," ) { @pe = add_family( 2, @pe ); }
            elsif ( $pe[0] eq "3family," ) { @pe = add_family( 3, @pe ); }
            elsif ( $pe[0] eq "family|r" ) { @pe = add_family_fix( @pe ); }
            elsif ( $pe[0] eq "'" ) { @pe = add_commaer( @pe ); }
            else
            {
                print "Dying ....\n";
                print "Error in tables! Unknown keyword at: @pe\n";
                die ;
            }
        }
    }
    close(TABLE);
}

#######################################################################
#########################                     #########################
#########################  ASSEMBLER PROPER   #########################
#########################                     #########################
#######################################################################

sub dump_code_space
{
    for( 0..$cp)
    {
        print $code[$_]," ";
    }
    print "\n";
}

# Assemble an opcode.
sub ass_opcode  # $mnemonic
{
    my $machinecode = $opcode{$_[0]};
    for ( 1..$opcode_length{$_[0]})
    {
        $code[$cp++] = $machinecode & 0xff ;
        $machinecode >>= 8;
    }
}

# Assemble a fixup.
sub ass_fixup   # $mnemonic
{
    my $machinecode = $fixup{$_[0]};
    my $bp = $cp;

    while ( $machinecode )
    {
        $code[--$bp] |= $machinecode & 0xff ;
        $machinecode >>= 8;
    }
}

# Assemble a comma.
sub ass_comma   # $mnemonic
{
    my $content = return_number;

    for ( 1..$commaer{$_[0]})
    {
        $code[$cp++] = $content & 0xff ;
        $content >>= 8;
    }
}

# Do a single pass, i.e. don't accomodate labels.
# Assume INFILE is open.
sub single_pass
{
    my $line_number=0;
    $cp = 0;

    LINE: while(<INFILE>)
    {
        split(' ',$_);

        # Within the assembler ``\ '' is the (Forth-compatible) comment.
        while(exists $_[0] and $_[0] ne "\\")
        {
            # Unbalanced left square bracket: a sib starter like [DI.
            if ( $_[0] =~ /^\[[^]]*$/ )
            {
                ass_fixup( "~SIB|" );
                ass_opcode( "SIB,," );
            };

            if    ( exists $opcode{ $_[0] } )    { ass_opcode( shift(@_) ); }
            elsif ( exists $fixup{ $_[0] } )     { ass_fixup(  shift(@_) ); }
            elsif ( exists $commaer{ $_[0] } )   { ass_comma(  shift(@_) ); }
            elsif ( $_[0] =~ /^[0-9A-F,]*$/ )    { hex_number( shift(@_) ); }
            else
            {
                print "Warning at line ", $.,
                    " : Can't understand input at : @_\n";
                next LINE;
            }
        }
    }
    close(TABLE);
}

# Output code in a convenient format for testing
# After testing it should be converted to intel Hex format.
sub output_code
{
    for (0..$#code)
    {
        print OUTFILE   sprintf("%02X", $code[$_]), ( $_%2 ? " " : "" ),  
                        ($_%0x10 == 0x0f ? "\n" : "" );
    }
    print OUTFILE   "\n";
}


get_file_handles();
get_tables;
single_pass;
output_code;

############################################################
# Left in for desperado debugging.
#print " --------- tables ------------\n";
#print "opcodes\n";
#foreach ( %opcode) {  print  $_,"\n"; };
#print " -----------------------------\n";
#print "fixups\n";
#foreach ( %fixup) {  print $_,"\n"; };
#print " -----------------------------\n";
#print "commaers\n";
#foreach ( %commaer) {  print $_,"\n"; };
#dump_code_space;
#dump_numbers;
